@@ -11,41 +11,34 @@
 -- License: MIT
 -- Maintainer: Jared Tobin <jared@ppad.tech>
 --
--- Constant-time choice.
+-- Primitives for constant-time choice.
+--
+-- The 'Choice' type encodes truthy and falsy values as unboxed 'Word#'
+-- bit masks.
+--
+-- Use the standard logical primitives ('or', 'and', 'xor', 'not', eq')
+-- to manipulate in-flight 'Choice' values. Use one of the selection
+-- functions to use a 'Choice' to select a value in constant time,
+-- or 'decide' to reduce a 'Choice' to a 'Bool' at the /end/ of a
+-- sensitive computation.
 
 module Data.Choice (
   -- * Choice
     Choice
+  , decide
   , true#
   , false#
-  , decide
   , to_word#
 
-  -- * MaybeWord#
-  , MaybeWord#(..)
-  , some_word#
-  , none_word#
-
-  -- * MaybeWide#
-  , MaybeWide#(..)
-  , some_wide#
-  , just_wide#
-  , none_wide#
-  , expect_wide#
-  , expect_wide_or#
-
   -- * Construction
-  , from_word_mask#
-  , from_word#
+  , from_full_mask#
+  , from_bit#
   , from_word_nonzero#
   , from_word_eq#
   , from_word_le#
   , from_word_lt#
   , from_word_gt#
 
-  , from_wide#
-  , from_wide_le#
-
   -- * Manipulation
   , or
   , and
@@ -72,82 +65,52 @@ import Prelude hiding (and, not, or)
 
 -- utilities ------------------------------------------------------------------
 
+type Limb2 = (# Word#, Word# #)
+
+type Limb4 = (# Word#, Word#, Word#, Word# #)
+
 -- wrapping negation
 neg_w# :: Word# -> Word#
 neg_w# w = Exts.plusWord# (Exts.not# w) 1##
 {-# INLINE neg_w# #-}
 
-hi# :: Word# -> (# Word#, Word# #)
+hi# :: Word# -> Limb2
 hi# w = (# 0##, w #)
 {-# INLINE hi# #-}
 
-lo# :: Word# -> (# Word#, Word# #)
+lo# :: Word# -> Limb2
 lo# w = (# w, 0## #)
 {-# INLINE lo# #-}
 
-not_w# :: (# Word#, Word# #) -> (# Word#, Word# #)
-not_w# (# a0, a1 #) = (# Exts.not# a0, Exts.not# a1 #)
-{-# INLINE not_w# #-}
-
-or_w# :: (# Word#, Word# #) -> (# Word#, Word# #) -> (# Word#, Word# #)
+or_w# :: Limb2 -> Limb2 -> Limb2
 or_w# (# a0, a1 #) (# b0, b1 #) = (# Exts.or# a0 b0, Exts.or# a1 b1 #)
 {-# INLINE or_w# #-}
 
-and_w# :: (# Word#, Word# #) -> (# Word#, Word# #) -> (# Word#, Word# #)
+and_w# :: Limb2 -> Limb2 -> Limb2
 and_w# (# a0, a1 #) (# b0, b1 #) = (# Exts.and# a0 b0, Exts.and# a1 b1 #)
 {-# INLINE and_w# #-}
 
-xor_w# :: (# Word#, Word# #) -> (# Word#, Word# #) -> (# Word#, Word# #)
+xor_w# :: Limb2 -> Limb2 -> Limb2
 xor_w# (# a0, a1 #) (# b0, b1 #) = (# Exts.xor# a0 b0, Exts.xor# a1 b1 #)
 {-# INLINE xor_w# #-}
 
--- subtract-with-borrow
-sub_b# :: Word# -> Word# -> Word# -> (# Word#, Word# #)
-sub_b# m n b =
-  let !(# d0, b0 #) = Exts.subWordC# m n
-      !(#  d, b1 #) = Exts.subWordC# d0 b
-      !c = Exts.int2Word# (Exts.orI# b0 b1)
-  in  (# d, c #)
-{-# INLINE sub_b# #-}
-
--- wide subtract-with-borrow
-sub_wb#
-  :: (# Word#, Word# #)
-  -> (# Word#, Word# #)
-  -> (# Word#, Word#, Word# #)
-sub_wb# (# a0, a1 #) (# b0, b1 #) =
-  let !(# s0, c0 #) = sub_b# a0 b0 0##
-      !(# s1, c1 #) = sub_b# a1 b1 c0
-  in  (# s0, s1, c1 #)
-{-# INLINE sub_wb# #-}
-
--- wide subtraction (wrapping)
-sub_w#
-  :: (# Word#, Word# #)
-  -> (# Word#, Word# #)
-  -> (# Word#, Word# #)
-sub_w# a b =
-  let !(# c0, c1, _ #) = sub_wb# a b
-  in  (# c0, c1 #)
-{-# INLINE sub_w# #-}
-
 -- choice ---------------------------------------------------------------------
 
 -- | Constant-time choice, encoded as a mask.
 --
---   Note that 'Choice' is defined as an unboxed newtype, and so a
+--   Note that 'Choice' is defined as an unlifted newtype, and so a
 --   'Choice' value cannot be bound at the top level. You should work
 --   with it locally in the context of a computation.
 --
---   It's safe to 'decide' a choice, reducing it to a 'Bool', at any
---   time, but the general encouraged pattern is to do that only at the
---   end of a computation.
+--   Use one of the selection functions to select a 'Choice' value in
+--   constant time, or 'decide' to reduce it to a 'Bool' at the /end/ of
+--   a sensitive computation.
 --
 --   >>> decide (or# (false# ()) (true# ()))
 --   True
 newtype Choice = Choice Word#
 
--- | Construct the falsy value.
+-- | Construct the falsy 'Choice'.
 --
 --   >>> decide (false# ())
 --   False
@@ -155,7 +118,7 @@ false# :: () -> Choice
 false# _ = Choice 0##
 {-# INLINE false# #-}
 
--- | Construct the truthy value.
+-- | Construct the truthy 'Choice'.
 --
 --   >>> decide (true# ())
 --   True
@@ -166,6 +129,13 @@ true# _ = case maxBound :: Word of
 
 -- | Decide a 'Choice' by reducing it to a 'Bool'.
 --
+--   The 'decide' function itself runs in constant time, but once
+--   it reduces a 'Choice' to a 'Bool', any subsequent branching on
+--   the result is liable to introduce variable-time behaviour.
+--
+--   You should 'decide' only at the /end/ of a computation, after all
+--   security-sensitive computations have been carried out.
+--
 --   >>> decide (true# ())
 --   True
 decide :: Choice -> Bool
@@ -173,91 +143,44 @@ decide (Choice c) = Exts.isTrue# (Exts.neWord# c 0##)
 {-# INLINE decide #-}
 
 -- | Convert a 'Choice' to an unboxed 'Word#'.
+--
+--   This essentially "unboxes" the 'Choice' for direct manipulation.
+--
+--   >>> import qualified GHC.Exts as Exts
+--   >>> Exts.isTrue# (Exts.eqWord# 0## (to_word# (false# ())))
+--   True
 to_word# :: Choice -> Word#
 to_word# (Choice c) = Exts.and# c 1##
 {-# INLINE to_word# #-}
 
--- constant time 'Maybe Word#'
-newtype MaybeWord# = MaybeWord# (# Word#, Choice #)
-
-some_word# :: Word# -> MaybeWord#
-some_word# w = MaybeWord# (# w, true# () #)
-{-# INLINE some_word# #-}
-
-none_word# :: Word# -> MaybeWord#
-none_word# w = MaybeWord# (# w, false# () #)
-{-# INLINE none_word# #-}
-
--- constant time 'Maybe (# Word#, Word# #)'
-newtype MaybeWide# = MaybeWide# (# (# Word#, Word# #), Choice #)
-
-just_wide# :: (# Word#, Word# #) -> Choice -> MaybeWide#
-just_wide# w c = MaybeWide# (# w, c #)
-{-# INLINE just_wide# #-}
-
-some_wide# :: (# Word#, Word# #) -> MaybeWide#
-some_wide# w = MaybeWide# (# w, true# () #)
-{-# INLINE some_wide# #-}
-
-none_wide# :: (# Word#, Word# #) -> MaybeWide#
-none_wide# w = MaybeWide# (# w, false# () #)
-{-# INLINE none_wide# #-}
-
-expect_wide# :: MaybeWide# -> String -> (# Word#, Word# #)
-expect_wide# (MaybeWide# (# w, Choice c #)) msg
-    | Exts.isTrue# (Exts.eqWord# c t#) = w
-    | otherwise = error $ "ppad-fixed (expect_wide#): " <> msg
-  where
-    !(Choice t#) = true# ()
-{-# INLINE expect_wide# #-}
-
-expect_wide_or# :: MaybeWide# -> (# Word#, Word# #) -> (# Word#, Word# #)
-expect_wide_or# (MaybeWide# (# w, Choice c #)) alt
-    | Exts.isTrue# (Exts.eqWord# c t#) = w
-    | otherwise = alt
-  where
-    !(Choice t#) = true# ()
-{-# INLINE expect_wide_or# #-}
-
 -- construction ---------------------------------------------------------------
 
--- | Construct a 'Choice' from an unboxed mask.
+-- | Construct a 'Choice' from an unboxed full-word mask.
 --
---   The input is /not/ checked.
+--   The input is /not/ checked to be a full-word mask.
 --
---   >>> decide (from_word_mask# 0##)
+--   >>> decide (from_full_mask# 0##)
 --   False
---   >>> decide (from_word_mask# 0xFFFFFFFFF_FFFFFFFF##)
+--   >>> decide (from_full_mask# 0xFFFFFFFFF_FFFFFFFF##)
 --   True
-from_word_mask# :: Word# -> Choice
-from_word_mask# w = Choice w
-{-# INLINE from_word_mask# #-}
+from_full_mask# :: Word# -> Choice
+from_full_mask# w = Choice w
+{-# INLINE from_full_mask# #-}
 
 -- | Construct a 'Choice' from an unboxed word, which should be either
 --   0## or 1##.
 --
---   The input is /not/ checked.
+--   The input is /not/ checked to be a bit.
 --
---   >>> decide (from_word# 1##)
+--   >>> decide (from_bit# 1##)
 --   True
-from_word# :: Word# -> Choice
-from_word# w = Choice (neg_w# w)
-{-# INLINE from_word# #-}
-
--- | Construct a 'Choice' from a two-limb word, constructing a mask from
---   the lower limb, which should be 0## or 1##.
---
---   The input is /not/ checked.
---
---   >>> decide (from_wide# (# 0##, 1## #))
---   False
-from_wide# :: (# Word#, Word# #) -> Choice
-from_wide# (# l, _ #) = from_word# l
-{-# INLINE from_wide# #-}
+from_bit# :: Word# -> Choice
+from_bit# w = Choice (neg_w# w)
+{-# INLINE from_bit# #-}
 
 -- | Construct a 'Choice' from a /nonzero/ unboxed word.
 --
---   The input is /not/ checked.
+--   The input is /not/ checked to be nonzero.
 --
 --   >>> decide (from_word_nonzero# 2##)
 --   True
@@ -266,7 +189,7 @@ from_word_nonzero# w =
   let !n = neg_w# w
       !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
       !v = Exts.uncheckedShiftRL# (Exts.or# w n) s
-  in  from_word# v
+  in  from_bit# v
 {-# INLINE from_word_nonzero# #-}
 
 -- | Construct a 'Choice' from an equality comparison.
@@ -280,7 +203,7 @@ from_word_eq# x y = case from_word_nonzero# (Exts.xor# x y) of
   Choice w -> Choice (Exts.not# w)
 {-# INLINE from_word_eq# #-}
 
--- | Construct a 'Choice from an at most comparison.
+-- | Construct a 'Choice from an at-most comparison.
 --
 --   >>> decide (from_word_le# 0## 1##)
 --   True
@@ -295,28 +218,9 @@ from_word_le# x y =
             (Exts.or# (Exts.not# x) y)
             (Exts.or# (Exts.xor# x y) (Exts.not# (Exts.minusWord# y x))))
           s
-  in  from_word# bit
+  in  from_bit# bit
 {-# INLINE from_word_le# #-}
 
--- | Construct a 'Choice' from an at most comparison on a two-limb
---   unboxed word.
---
---   >>> decide (from_wide_le# (# 0##, 0## #) (# 1##, 0## #))
---   True
---   >>> decide (from_wide_le# (# 1##, 0## #) (# 1##, 0## #))
---   True
-from_wide_le# :: (# Word#, Word# #) -> (# Word#, Word# #) -> Choice
-from_wide_le# x y =
-  let !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
-      !mask =
-        (and_w#
-          (or_w# (not_w# x) y)
-          (or_w# (xor_w# x y) (not_w# (sub_w# y x))))
-      !bit = case mask of
-        (# l, _ #) -> Exts.uncheckedShiftRL# l s
-  in  from_word# bit
-{-# INLINE from_wide_le# #-}
-
 -- | Construct a 'Choice' from a less-than comparison.
 --
 --   >>> decide (from_word_lt# 0## 1##)
@@ -332,7 +236,7 @@ from_word_lt# x y =
             (Exts.and# (Exts.not# x) y)
             (Exts.and# (Exts.or# (Exts.not# x) y) (Exts.minusWord# x y)))
           s
-  in  from_word# bit
+  in  from_bit# bit
 {-# INLINE from_word_lt# #-}
 
 -- | Construct a 'Choice' from a greater-than comparison.
@@ -348,59 +252,82 @@ from_word_gt# x y = from_word_lt# y x
 -- manipulation ---------------------------------------------------------------
 
 -- | Logically negate a 'Choice'.
+--
+--   >>> decide (not (true# ()))
+--   False
+--   >>> decide (not (false# ()))
+--   True
 not :: Choice -> Choice
 not (Choice w) = Choice (Exts.not# w)
 {-# INLINE not #-}
 
 -- | Logical disjunction on 'Choice' values.
+--
+--   >>> decide (or (true# ()) (false# ()))
+--   True
 or :: Choice -> Choice -> Choice
 or (Choice w0) (Choice w1) = Choice (Exts.or# w0 w1)
 {-# INLINE or #-}
 
 -- | Logical conjunction on 'Choice' values.
+--
+--   >>> decide (and (true# ()) (false# ()))
+--   False
 and :: Choice -> Choice -> Choice
 and (Choice w0) (Choice w1) = Choice (Exts.and# w0 w1)
 {-# INLINE and #-}
 
 -- | Logical inequality on 'Choice' values.
+--
+--   >>> decide (xor (true# ()) (false# ()))
+--   True
 xor :: Choice -> Choice -> Choice
 xor (Choice w0) (Choice w1) = Choice (Exts.xor# w0 w1)
 {-# INLINE xor #-}
 
 -- | Logical inequality on 'Choice' values.
+--
+--   >>> decide (ne (true# ()) (false# ()))
+--   True
 ne :: Choice -> Choice -> Choice
 ne c0 c1 = xor c0 c1
 {-# INLINE ne #-}
 
 -- | Logical equality on 'Choice' values.
+--
+--   >>> decide (eq (true# ()) (false# ()))
+--   False
 eq :: Choice -> Choice -> Choice
 eq c0 c1 = not (ne c0 c1)
 {-# INLINE eq #-}
 
 -- constant-time selection ----------------------------------------------------
 
--- | Select an unboxed word, given a 'Choice'.
+-- | Select an unboxed word without branching, given a 'Choice'.
+--
+--   >>> let w = C.select_word# 0## 1## (C.true# ()) in GHC.Word.W# w
+--   1
 select_word# :: Word# -> Word# -> Choice -> Word#
 select_word# a b (Choice c) = Exts.xor# a (Exts.and# c (Exts.xor# a b))
 {-# INLINE select_word# #-}
 
--- | Select an unboxed two-limb word, given a 'Choice'.
+-- | Select an unboxed two-limb word without branching, given a 'Choice'.
 select_wide#
-  :: (# Word#, Word# #)
-  -> (# Word#, Word# #)
+  :: Limb2
+  -> Limb2
   -> Choice
-  -> (# Word#, Word# #)
+  -> Limb2
 select_wide# a b (Choice w) =
   let !mask = or_w# (hi# w) (lo# w)
   in  xor_w# a (and_w# mask (xor_w# a b))
 {-# INLINE select_wide# #-}
 
--- | Select an unboxed four-limb word, given a 'Choice'.
+-- | Select an unboxed four-limb word without branching, given a 'Choice'.
 select_wider#
-  :: (# Word#, Word#, Word#, Word# #)
-  -> (# Word#, Word#, Word#, Word# #)
+  :: Limb4
+  -> Limb4
   -> Choice
-  -> (# Word#, Word#, Word#, Word# #)
+  -> Limb4
 select_wider# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) (Choice w) =
   let !w0 = Exts.xor# a0 (Exts.and# w (Exts.xor# a0 b0))
       !w1 = Exts.xor# a1 (Exts.and# w (Exts.xor# a1 b1))
@@ -428,8 +355,8 @@ eq_word# a b =
 --   >>> decide (eq_wide (# 0##, 0## #) (# 0##, 0## #))
 --   True
 eq_wide#
-  :: (# Word#, Word# #)
-  -> (# Word#, Word# #)
+  :: Limb2
+  -> Limb2
   -> Choice
 eq_wide# (# a0, a1 #) (# b0, b1 #) =
   let !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
@@ -443,8 +370,8 @@ eq_wide# (# a0, a1 #) (# b0, b1 #) =
 --   >>> let zero = (# 0##, 0##, 0##, 0## #) in decide (eq_wider# zero zero)
 --   True
 eq_wider#
-  :: (# Word#, Word#, Word#, Word# #)
-  -> (# Word#, Word#, Word#, Word# #)
+  :: Limb4
+  -> Limb4
   -> Choice
 eq_wider# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
   let !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
