@@ -53,9 +53,11 @@ module Data.Word.Wider (
   , shr_limb#
   , shl_limb#
   , and
-  , and_w#
+  , and#
   , or
-  , or_w#
+  , or#
+  , xor
+  , xor#
   , not
   , not#
 
@@ -97,17 +99,17 @@ fi = fromIntegral
 
 -- wider words ----------------------------------------------------------------
 
-pattern Limb4
-  :: Word# -> Word# -> Word# -> Word#
-  -> (# Limb, Limb, Limb, Limb #)
-pattern Limb4 w0 w1 w2 w3 = (# Limb w0, Limb w1, Limb w2, Limb w3 #)
-{-# COMPLETE Limb4 #-}
+type Limb4 = (# Limb, Limb, Limb, Limb #)
+
+pattern L4 :: Word# -> Word# -> Word# -> Word# -> Limb4
+pattern L4 w0 w1 w2 w3 = (# Limb w0, Limb w1, Limb w2, Limb w3 #)
+{-# COMPLETE L4 #-}
 
 -- | Little-endian wider words, consisting of four 'Limbs'.
 --
 --   >>> 1 :: Wider
 --   1
-data Wider = Wider !(# Limb, Limb, Limb, Limb #)
+data Wider = Wider !Limb4
 
 instance Show Wider where
   show = show . from_vartime
@@ -121,12 +123,12 @@ instance Num Wider where
   (*) = mul
   abs = id
   fromInteger = to_vartime
-  negate w = add (not w) (Wider (Limb4 1## 0## 0## 0##))
+  negate w = add (not w) (Wider (L4 1## 0## 0## 0##))
   signum (Wider (# l0, l1, l2, l3 #)) =
     let !(Limb l) = l0 `L.or#` l1 `L.or#` l2 `L.or#` l3
         !n = C.from_word_nonzero# l
         !b = C.to_word# n
-    in  Wider (Limb4 b 0## 0## 0##)
+    in  Wider (L4 b 0## 0## 0##)
 
 instance NFData Wider where
   rnf (Wider a) = case a of
@@ -135,12 +137,12 @@ instance NFData Wider where
 -- comparison -----------------------------------------------------------------
 
 eq#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
+  :: Limb4
+  -> Limb4
   -> C.Choice
 eq# a b =
-  let !(Limb4 a0 a1 a2 a3) = a
-      !(Limb4 b0 b1 b2 b3) = b
+  let !(L4 a0 a1 a2 a3) = a
+      !(L4 b0 b1 b2 b3) = b
   in  C.eq_wider# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #)
 {-# INLINE eq# #-}
 
@@ -158,14 +160,15 @@ eq_vartime a b =
       && (L.eq_vartime# a1 b1)
       && (L.eq_vartime# a2 b2)
       && (L.eq_vartime# a3 b3)
+{-# INLINABLE eq_vartime #-}
 
 lt#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
+  :: Limb4
+  -> Limb4
   -> C.Choice
 lt# a b =
   let !(# _, Limb bor #) = sub_b# a b
-  in  C.from_word_mask# bor
+  in  C.from_full_mask# bor
 {-# INLINE lt# #-}
 
 -- | Constant-time less-than comparison between 'Wider' values.
@@ -177,14 +180,15 @@ lt# a b =
 --   False
 lt :: Wider -> Wider -> C.Choice
 lt (Wider a) (Wider b) = lt# a b
+{-# INLINABLE lt #-}
 
 gt#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
+  :: Limb4
+  -> Limb4
   -> C.Choice
 gt# a b =
   let !(# _, Limb bor #) = sub_b# b a
-  in  C.from_word_mask# bor
+  in  C.from_full_mask# bor
 {-# INLINE gt# #-}
 
 -- | Constant-time greater-than comparison between 'Wider' values.
@@ -196,10 +200,11 @@ gt# a b =
 --   True
 gt :: Wider -> Wider -> C.Choice
 gt (Wider a) (Wider b) = gt# a b
+{-# INLINABLE gt #-}
 
 cmp#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
+  :: Limb4
+  -> Limb4
   -> Int#
 cmp# (# l0, l1, l2, l3 #) (# r0, r1, r2, r3 #) =
   let !(# w0, b0 #) = L.sub_b# r0 l0 (Limb 0##)
@@ -241,8 +246,8 @@ cmp_vartime (Wider a) (Wider b) = case cmp# a b of
 --   >>> wider 1 0 0 0
 --   1
 wider :: Word -> Word -> Word -> Word -> Wider
-wider (W# w0) (W# w1) (W# w2) (W# w3) = Wider
-  (# Limb w0, Limb w1, Limb w2, Limb w3 #)
+wider (W# w0) (W# w1) (W# w2) (W# w3) = Wider (L4 w0 w1 w2 w3)
+{-# INLINABLE wider #-}
 
 -- | Convert an 'Integer' to a 'Wider' word.
 --
@@ -256,34 +261,34 @@ to_vartime n =
       !(W# w1) = fi ((n .>>. size) .&. mask)
       !(W# w2) = fi ((n .>>. (2 * size)) .&. mask)
       !(W# w3) = fi ((n .>>. (3 * size)) .&. mask)
-  in  Wider (# Limb w0, Limb w1, Limb w2, Limb w3 #)
+  in  Wider (L4 w0 w1 w2 w3)
+{-# INLINABLE to_vartime #-}
 
 -- | Convert a 'Wider' word to an 'Integer'.
 --
 --   >>> from_vartime 1
 --   1
 from_vartime :: Wider -> Integer
-from_vartime (Wider (# Limb w0, Limb w1, Limb w2, Limb w3 #)) =
+from_vartime (Wider (L4 w0 w1 w2 w3)) =
         fi (W# w3) .<<. (3 * size)
     .|. fi (W# w2) .<<. (2 * size)
     .|. fi (W# w1) .<<. size
     .|. fi (W# w0)
   where
     !size = B.finiteBitSize (0 :: Word)
+{-# INLINABLE from_vartime #-}
 
 -- constant-time selection-----------------------------------------------------
 
 select#
-  :: (# Limb, Limb, Limb, Limb #) -- ^ a
-  -> (# Limb, Limb, Limb, Limb #) -- ^ b
-  -> C.Choice                     -- ^ c
-  -> (# Limb, Limb, Limb, Limb #) -- ^ result
-select# a b c =
-  let !(# Limb a0, Limb a1, Limb a2, Limb a3 #) = a
-      !(# Limb b0, Limb b1, Limb b2, Limb b3 #) = b
-      !(# w0, w1, w2, w3 #) =
+  :: Limb4    -- ^ a
+  -> Limb4    -- ^ b
+  -> C.Choice -- ^ c
+  -> Limb4    -- ^ result
+select# (L4 a0 a1 a2 a3) (L4 b0 b1 b2 b3) c =
+  let !(# w0, w1, w2, w3 #) =
         C.select_wider# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) c
-  in  (# Limb w0, Limb w1, Limb w2, Limb w3 #)
+  in  L4 w0 w1 w2 w3
 {-# INLINE select# #-}
 
 -- | Return a if c is truthy, otherwise return b.
@@ -297,12 +302,13 @@ select
   -> C.Choice -- ^ c
   -> Wider    -- ^ result
 select (Wider a) (Wider b) c = Wider (select# a b c)
+{-# INLINABLE select #-}
 
 -- bit manipulation -----------------------------------------------------------
 
 shr1_c#
-  :: (# Limb, Limb, Limb, Limb #)                 -- ^ argument
-  -> (# (# Limb, Limb, Limb, Limb #), C.Choice #) -- ^ result, carry
+  :: Limb4                 -- ^ argument
+  -> (# Limb4, C.Choice #) -- ^ result, carry
 shr1_c# (# w0, w1, w2, w3 #) =
   let !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
       !(# s3, c3 #) = (# L.shr# w3 1#, L.shl# w3 s #)
@@ -314,7 +320,7 @@ shr1_c# (# w0, w1, w2, w3 #) =
       !(# s0, c0 #) = (# L.shr# w0 1#, L.shl# w0 s #)
       !r0           = L.or# s0 c1
       !(Limb w)     = L.shr# c0 s
-  in  (# (# r0, r1, r2, r3 #), C.from_word# w #)
+  in  (# (# r0, r1, r2, r3 #), C.from_bit# w #)
 {-# INLINE shr1_c# #-}
 
 -- | Constant-time 1-bit shift-right with carry, with a 'Choice'
@@ -323,6 +329,7 @@ shr1_c :: Wider -> (# Wider, C.Choice #)
 shr1_c (Wider w) =
   let !(# r, c #) = shr1_c# w
   in  (# Wider r, c #)
+{-# INLINABLE shr1_c #-}
 
 -- | Constant-time 1-bit shift-right.
 --
@@ -334,10 +341,11 @@ shr1 :: Wider -> Wider
 shr1 (Wider w) =
   let !(# r, _ #) = shr1_c# w
   in  Wider r
+{-# INLINABLE shr1 #-}
 
 shl1_c#
-  :: (# Limb, Limb, Limb, Limb #)                 -- ^ argument
-  -> (# (# Limb, Limb, Limb, Limb #), C.Choice #) -- ^ result, carry
+  :: Limb4                 -- ^ argument
+  -> (# Limb4, C.Choice #) -- ^ result, carry
 shl1_c# (# w0, w1, w2, w3 #) =
   let !s = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
       !(# s0, c0 #) = (# L.shl# w0 1#, L.shr# w0 s #)
@@ -349,7 +357,7 @@ shl1_c# (# w0, w1, w2, w3 #) =
       !(# s3, c3 #) = (# L.shl# w3 1#, L.shr# w3 s #)
       !r3           = L.or# s3 c2
       !(Limb w)     = L.shl# c3 s
-  in  (# (# r0, r1, r2, r3 #), C.from_word# w #)
+  in  (# (# r0, r1, r2, r3 #), C.from_bit# w #)
 {-# INLINE shl1_c# #-}
 
 -- | Constant-time 1-bit shift-left with carry, with a 'Choice' indicating
@@ -358,6 +366,7 @@ shl1_c :: Wider -> (# Wider, C.Choice #)
 shl1_c (Wider w) =
   let !(# r, c #) = shl1_c# w
   in  (# Wider r, c #)
+{-# INLINABLE shl1_c #-}
 
 -- | Constant-time 1-bit shift-left.
 --
@@ -369,14 +378,14 @@ shl1 :: Wider -> Wider
 shl1 (Wider w) =
   let !(# r, _ #) = shl1_c# w
   in  Wider r
+{-# INLINABLE shl1 #-}
 
 shr_limb#
-  :: (# Limb, Limb, Limb, Limb #)
+  :: Limb4
   -> Int#
-  -> (# (# Limb, Limb, Limb, Limb #), Limb #)
+  -> (# Limb4, Limb #)
 shr_limb# (# a0, a1, a2, a3 #) rs =
-  let !size = case B.finiteBitSize (0 :: Word) of I# m -> m
-      !ls = size Exts.-# rs
+  let !ls = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# rs
       !(# l3, c3 #) = (# L.shr# a3 rs, L.shl# a3 ls #)
       !(# l2, c2 #) = (# L.or# (L.shr# a2 rs) c3, L.shl# a2 ls #)
       !(# l1, c1 #) = (# L.or# (L.shr# a1 rs) c2, L.shl# a1 ls #)
@@ -397,14 +406,14 @@ shr_limb
 shr_limb (Wider w) (I# s) =
   let !(# r, _ #) = shr_limb# w s
   in  Wider r
+{-# INLINABLE shr_limb #-}
 
 shl_limb#
-  :: (# Limb, Limb, Limb, Limb #)
+  :: Limb4
   -> Int#
-  -> (# (# Limb, Limb, Limb, Limb #), Limb #)
+  -> (# Limb4, Limb #)
 shl_limb# (# a0, a1, a2, a3 #) ls =
-  let !size = case B.finiteBitSize (0 :: Word) of I# m -> m
-      !rs = size Exts.-# ls
+  let !rs = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# ls
       !(# l0, c0 #) = (# L.shl# a0 ls, L.shr# a0 rs #)
       !(# l1, c1 #) = (# L.or# (L.shl# a1 ls) c0, L.shr# a1 rs #)
       !(# l2, c2 #) = (# L.or# (L.shl# a2 ls) c1, L.shr# a2 rs #)
@@ -427,14 +436,15 @@ shl_limb
 shl_limb (Wider w) (I# s) =
   let !(# r, _ #) = shl_limb# w s
   in  Wider r
+{-# INLINABLE shl_limb #-}
 
-and_w#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
-and_w# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
+and#
+  :: Limb4
+  -> Limb4
+  -> Limb4
+and# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
   (# L.and# a0 b0, L.and# a1 b1, L.and# a2 b2, L.and# a3 b3 #)
-{-# INLINE and_w# #-}
+{-# INLINE and# #-}
 
 -- | Binary /and/.
 --
@@ -446,15 +456,16 @@ and
   :: Wider -- ^ a
   -> Wider -- ^ b
   -> Wider -- ^ a & b
-and (Wider a) (Wider b) = Wider (and_w# a b)
+and (Wider a) (Wider b) = Wider (and# a b)
+{-# INLINABLE and #-}
 
-or_w#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
-or_w# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
+or#
+  :: Limb4
+  -> Limb4
+  -> Limb4
+or# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
   (# L.or# a0 b0, L.or# a1 b1, L.or# a2 b2, L.or# a3 b3 #)
-{-# INLINE or_w# #-}
+{-# INLINE or# #-}
 
 -- | Binary /or/.
 --
@@ -466,11 +477,33 @@ or
   :: Wider -- ^ a
   -> Wider -- ^ b
   -> Wider -- ^ a | b
-or (Wider a) (Wider b) = Wider (or_w# a b)
+or (Wider a) (Wider b) = Wider (or# a b)
+{-# INLINABLE or #-}
+
+xor#
+  :: Limb4
+  -> Limb4
+  -> Limb4
+xor# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
+  (# L.xor# a0 b0, L.xor# a1 b1, L.xor# a2 b2, L.xor# a3 b3 #)
+{-# INLINE xor# #-}
+
+-- | Binary /xor/.
+--
+--   >>> xor 1 1
+--   0
+--   >>> xor 1 0
+--   1
+xor
+  :: Wider -- ^ a
+  -> Wider -- ^ b
+  -> Wider -- ^ a ^ b
+xor (Wider a) (Wider b) = Wider (xor# a b)
+{-# INLINABLE xor #-}
 
 not#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
+  :: Limb4
+  -> Limb4
 not# (# l0, l1, l2, l3 #) = (# L.not# l0, L.not# l1, L.not# l2, L.not# l3 #)
 {-# INLINE not# #-}
 
@@ -484,13 +517,14 @@ not
   :: Wider -- ^ value
   -> Wider -- ^ not value
 not (Wider w) = Wider (not# w)
+{-# INLINABLE not #-}
 
 -- addition, subtraction ------------------------------------------------------
 
 add_o#
-  :: (# Limb, Limb, Limb, Limb #)             -- ^ augend
-  -> (# Limb, Limb, Limb, Limb #)             -- ^ addend
-  -> (# (# Limb, Limb, Limb, Limb #), Limb #) -- ^ (# sum, carry bit #)
+  :: Limb4             -- ^ augend
+  -> Limb4             -- ^ addend
+  -> (# Limb4, Limb #) -- ^ (# sum, carry bit #)
 add_o# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
   let !(# s0, c0 #) = L.add_o# a0 b0
       !(# s1, c1 #) = L.add_c# a1 b1 c0
@@ -513,11 +547,12 @@ add_o
 add_o (Wider a) (Wider b) =
   let !(# s, Limb c #) = add_o# a b
   in  (Wider s, W# c)
+{-# INLINABLE add_o #-}
 
 add_w#
-  :: (# Limb, Limb, Limb, Limb #) -- ^ augend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ addend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ sum
+  :: Limb4 -- ^ augend
+  -> Limb4 -- ^ addend
+  -> Limb4 -- ^ sum
 add_w# a b =
   let !(# c, _ #) = add_o# a b
   in  c
@@ -538,10 +573,10 @@ add (Wider a) (Wider b) = Wider (add_w# a b)
 {-# INLINE add #-}
 
 add_mod#
-  :: (# Limb, Limb, Limb, Limb #) -- ^ augend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ addend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ modulus
-  -> (# Limb, Limb, Limb, Limb #) -- ^ sum
+  :: Limb4 -- ^ augend
+  -> Limb4 -- ^ addend
+  -> Limb4 -- ^ modulus
+  -> Limb4 -- ^ sum
 add_mod# a b m =
   let !(# w, c #) = add_o# a b
   in  sub_mod_c# w c m m
@@ -562,11 +597,12 @@ add_mod
   -> Wider -- ^ modulus
   -> Wider -- ^ sum
 add_mod (Wider a) (Wider b) (Wider m) = Wider (add_mod# a b m)
+{-# INLINABLE add_mod #-}
 
 sub_b#
-  :: (# Limb, Limb, Limb, Limb #)              -- ^ minuend
-  -> (# Limb, Limb, Limb, Limb #)              -- ^ subtrahend
-  -> (# (# Limb, Limb, Limb, Limb #), Limb #) -- ^ (# diff, borrow mask #)
+  :: Limb4              -- ^ minuend
+  -> Limb4              -- ^ subtrahend
+  -> (# Limb4, Limb #) -- ^ (# diff, borrow mask #)
 sub_b# (# a0, a1, a2, a3 #) (# b0, b1, b2, b3 #) =
   let !(# s0, c0 #) = L.sub_b# a0 b0 (Limb 0##)
       !(# s1, c1 #) = L.sub_b# a1 b1 c0
@@ -589,6 +625,7 @@ sub_b
 sub_b (Wider l) (Wider r) =
   let !(# d, Limb b #) = sub_b# l r
   in  (Wider d, W# b)
+{-# INLINABLE sub_b #-}
 
 -- | Wrapping subtraction, computing 'a - b' and returning the
 --   difference.
@@ -607,12 +644,13 @@ sub
 sub (Wider a) (Wider b) =
   let !(# d, _ #) = sub_b# a b
   in  Wider d
+{-# INLINABLE sub #-}
 
 sub_mod#
-  :: (# Limb, Limb, Limb, Limb #) -- ^ minuend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ subtrahend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ modulus
-  -> (# Limb, Limb, Limb, Limb #) -- ^ difference
+  :: Limb4 -- ^ minuend
+  -> Limb4 -- ^ subtrahend
+  -> Limb4 -- ^ modulus
+  -> Limb4 -- ^ difference
 sub_mod# a b (# p0, p1, p2, p3 #) =
   let !(# o, m #) = sub_b# a b
       !ba = (# L.and# p0 m, L.and# p1 m, L.and# p2 m, L.and# p3 m #)
@@ -634,14 +672,15 @@ sub_mod
   -> Wider
   -> Wider
 sub_mod (Wider a) (Wider b) (Wider p) = Wider (sub_mod# a b p)
+{-# INLINABLE sub_mod #-}
 
 -- | Modular subtraction with carry. Computes (# a, c #) - b mod m.
 sub_mod_c#
-  :: (# Limb, Limb, Limb, Limb #) -- ^ minuend
+  :: Limb4 -- ^ minuend
   -> Limb                         -- ^ carry bit
-  -> (# Limb, Limb, Limb, Limb #) -- ^ subtrahend
-  -> (# Limb, Limb, Limb, Limb #) -- ^ modulus
-  -> (# Limb, Limb, Limb, Limb #) -- ^ difference
+  -> Limb4 -- ^ subtrahend
+  -> Limb4 -- ^ modulus
+  -> Limb4 -- ^ difference
 sub_mod_c# a c b (# p0, p1, p2, p3 #) =
   let !(# (# o0, o1, o2, o3 #), bb #) = sub_b# a b
       !(# _, m #) = L.sub_b# c (Limb 0##) bb
@@ -652,9 +691,9 @@ sub_mod_c# a c b (# p0, p1, p2, p3 #) =
 -- multiplication -------------------------------------------------------------
 
 mul_c#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# Limb, Limb, Limb, Limb #)
-  -> (# (# Limb, Limb, Limb, Limb #), (# Limb, Limb, Limb, Limb #) #)
+  :: Limb4
+  -> Limb4
+  -> (# Limb4, Limb4 #)
 mul_c# (# x0, x1, x2, x3 #) (# y0, y1, y2, y3 #) =
   let !(# z0, c0_0 #)   = L.mac# x0 y0 (Limb 0##) (Limb 0##)
       !(# s1_0, c1_0 #) = L.mac# x0 y1 (Limb 0##) c0_0
@@ -700,6 +739,7 @@ mul_c
 mul_c (Wider a) (Wider b) =
   let !(# l, h #) = mul_c# a b
   in  (Wider l, Wider h)
+{-# INLINABLE mul_c #-}
 
 -- | Wrapping multiplication.
 --
@@ -717,10 +757,11 @@ mul
 mul (Wider a) (Wider b) =
   let !(# l, _ #) = mul_c# a b
   in  Wider l
+{-# INLINABLE mul #-}
 
 sqr#
-  :: (# Limb, Limb, Limb, Limb #)
-  -> (# (# Limb, Limb, Limb, Limb #), (# Limb, Limb, Limb, Limb #) #)
+  :: Limb4
+  -> (# Limb4, Limb4 #)
 sqr# (# x0, x1, x2, x3 #) =
   let !sh = case B.finiteBitSize (0 :: Word) of I# m -> m Exts.-# 1#
       !(# q1_0, c1_0 #)  = L.mac# x1 x0 (Limb 0##) (Limb 0##)
@@ -760,9 +801,12 @@ sqr :: Wider -> (Wider, Wider)
 sqr (Wider w) =
   let !(# l, h #) = sqr# w
   in  (Wider l, Wider h)
+{-# INLINABLE sqr #-}
 
-odd# :: (# Limb, Limb, Limb, Limb #) -> C.Choice
-odd# (# Limb w, _, _, _ #) = C.from_word# (Exts.and# w 1##)
+odd# :: Limb4 -> C.Choice
+odd# (# l, _, _, _ #) =
+  let !(Limb w) = L.and# l (Limb 1##)
+  in  C.from_bit# w
 {-# INLINE odd# #-}
 
 -- | Check if a 'Wider' is odd, returning a 'Choice'.
@@ -770,4 +814,5 @@ odd
   :: Wider
   -> C.Choice
 odd (Wider w) = odd# w
+{-# INLINABLE odd #-}
 
